package caps

import (
	"errors"
	"sort"
)

type Capabilities map[string]int

const (
	CAP_USER_CREATE   string = "USER_CREATE"
	CAP_USER_DESTROY  string = "USER_DESTROY"
	CAP_USER_SET_CAPS string = "USER_SET_CAPS"
	CAP_USER_UNLOCK   string = "USER_UNLOCK"
	CAP_USER_RESETPWD string = "USER_RESETPWD"
	CAP_USER_LIST     string = "USER_LIST"

	CAP_VM_LIST           string = "VM_LIST"
	CAP_VM_LIST_ANY       string = "VM_LIST_ANY"
	CAP_VM_ATTACH         string = "VM_ATTACH"
	CAP_VM_ATTACH_ANY     string = "VM_ATTACH_ANY"
	CAP_VM_START          string = "VM_START"
	CAP_VM_START_ANY      string = "VM_START_ANY"
	CAP_VM_STOP           string = "VM_STOP"
	CAP_VM_STOP_ANY       string = "VM_STOP_ANY"
	CAP_VM_REBOOT         string = "VM_REBOOT"
	CAP_VM_REBOOT_ANY     string = "VM_REBOOT_ANY"
	CAP_VM_CREATE         string = "VM_CREATE"
	CAP_VM_DESTROY        string = "VM_DESTROY"
	CAP_VM_DESTROY_ANY    string = "VM_DESTROY_ANY"
	CAP_VM_EDIT           string = "VM_EDIT"
	CAP_VM_EDIT_ANY       string = "VM_EDIT_ANY"
	CAP_VM_SET_ACCESS     string = "VM_SET_ACCESS"
	CAP_VM_SET_ACCESS_ANY string = "VM_SET_ACCESS_ANY"
	CAP_VM_READFS         string = "VM_READFS"
	CAP_VM_READFS_ANY     string = "VM_READFS_ANY"
	CAP_VM_WRITEFS        string = "VM_WRITEFS"
	CAP_VM_WRITEFS_ANY    string = "VM_WRITEFS_ANY"

	CAP_EXAM_ATTACH string = "EXAM_ATTACH"

	CAP_TPL_CREATE      string = "TPL_CREATE"
	CAP_TPL_EDIT        string = "TPL_EDIT"
	CAP_TPL_EDIT_ANY    string = "TPL_EDIT_ANY"
	CAP_TPL_LIST        string = "TPL_LIST"
	CAP_TPL_LIST_ANY    string = "TPL_LIST_ANY"
	CAP_TPL_DESTROY     string = "TPL_DESTROY"
	CAP_TPL_DESTROY_ANY string = "TPL_DESTROY_ANY"
	CAP_TPL_READFS      string = "TPL_READFS"
	CAP_TPL_READFS_ANY  string = "TPL_READFS_ANY"
)

// Capabilities stored in the user config
var userCaps = Capabilities{
	CAP_USER_CREATE:   1,
	CAP_USER_DESTROY:  1,
	CAP_USER_SET_CAPS: 1,
	CAP_USER_UNLOCK:   1,
	CAP_USER_RESETPWD: 1,
	CAP_USER_LIST:     1,

	CAP_VM_CREATE:         1,
	CAP_VM_DESTROY_ANY:    1,
	CAP_VM_EDIT_ANY:       1,
	CAP_VM_START_ANY:      1,
	CAP_VM_STOP_ANY:       1,
	CAP_VM_REBOOT_ANY:     1,
	CAP_VM_LIST_ANY:       1,
	CAP_VM_ATTACH_ANY:     1,
	CAP_VM_SET_ACCESS:     1,
	CAP_VM_SET_ACCESS_ANY: 1,
	CAP_VM_READFS_ANY:     1,
	CAP_VM_WRITEFS_ANY:    1,

	CAP_EXAM_ATTACH: 1,

	CAP_TPL_CREATE:      1,
	CAP_TPL_EDIT:        1,
	CAP_TPL_EDIT_ANY:    1,
	CAP_TPL_DESTROY:     1,
	CAP_TPL_DESTROY_ANY: 1,
	CAP_TPL_LIST:        1,
	CAP_TPL_LIST_ANY:    1,
	CAP_TPL_READFS:      1,
	CAP_TPL_READFS_ANY:  1,
}

// Capabilities stored in the VM config (access)
var vmAccessCaps = Capabilities{
	CAP_VM_SET_ACCESS: 1,
	CAP_VM_DESTROY:    1,
	CAP_VM_EDIT:       1,
	CAP_VM_START:      1,
	CAP_VM_STOP:       1,
	CAP_VM_REBOOT:     1,
	CAP_VM_LIST:       1,
	CAP_VM_ATTACH:     1,
	CAP_VM_READFS:     1,
	CAP_VM_WRITEFS:    1,
}

// Returns true if the string cap matches a user capability
func IsUserCapValid(cap string) bool {
	_, exists := userCaps[cap]
	return exists
}

// Returns true if the string cap matches a VM access capability
func IsVMAccessCapValid(cap string) bool {
	_, exists := vmAccessCaps[cap]
	return exists
}

// Validates all user capabilities.
func ValidateUserCaps(caps Capabilities) error {
	for cap := range caps {
		if !IsUserCapValid(cap) {
			return errors.New("Invalid capability: " + cap)
		}
	}
	return nil
}

// Validates all VM access capabilities.
func ValidateVMAccessCaps(caps Capabilities) error {
	for cap := range caps {
		if !IsVMAccessCapValid(cap) {
			return errors.New("Invalid capability: " + cap)
		}
	}
	return nil
}

func GetUserCapsNames() []string {
	caps := []string{}
	for key := range userCaps {
		caps = append(caps, key)
	}

	// Sort caps by name
	sort.Slice(caps, func(i, j int) bool {
		return caps[i] < caps[j]
	})

	return caps
}

func GetVMAccessCapsNames() []string {
	caps := []string{}
	for key := range vmAccessCaps {
		caps = append(caps, key)
	}

	// Sort caps by name
	sort.Slice(caps, func(i, j int) bool {
		return caps[i] < caps[j]
	})

	return caps
}
